//
//  Node.m
//  OutlineMe
//
//  Created by John Schilling on 11/11/04.
//  Copyright 2004 John Schilling. All rights reserved.
//

#import "Node.h"
#import "AppController.h"



@implementation Node

#pragma mark
#pragma mark PRIVATE INSTANCE METHODS
#pragma mark

- (id)initWithDictionary:(NSDictionary *)aDict 
                        delegate:(id)delegate 
                        dataFolderPath:(NSString *)dataFolderPath
                        overrideDataMismatch:(BOOL)overrideDataMismatch
{
    self = [super init];
    
    if (self) {
        
        _lastUpdateString           = [[NSMutableString alloc] init];
        _nextUpdateString           = [[NSMutableString alloc] init];
        _shortLastUpdateString      = [[NSMutableString alloc] init];
        _shortNextUpdateString      = [[NSMutableString alloc] init];
        _statusMessage              = [[NSMutableString alloc] init];
        _longStatusMessage          = [[NSMutableString alloc] init];
        
        NSString *sourceFileName    = [aDict objectForKey:@"SourceFileName"];
        NSString *url               = [aDict objectForKey:@"URL"];
        
        _destFileName               = [[aDict objectForKey:@"DestFileName"] retain];
        _itemName                   = [[aDict objectForKey:@"NodeName"] retain];
        _nodeTitle                  = [[aDict objectForKey:@"NodeTitle"] retain];
        _itemDescription            = [[aDict objectForKey:@"Description"] retain];
        
        _delegate                   = delegate;
        _dict                       = [[NSDictionary dictionaryWithDictionary:aDict] retain];
        _dataFolderPath             = [dataFolderPath retain];
        _sourceURL                  = [[url stringByAppendingString:[aDict objectForKey:@"SourceFileName"]] retain];
        _filePath                   = [[_dataFolderPath stringByAppendingPathComponent:_destFileName] retain];
        _fileType                   = [[_dict objectForKey:@"FileType"] intValue];
        
        _overrideDataMismatch       = overrideDataMismatch; /* Used for Aurora Alerts */
        
        NSString *lastUpdated = [[NSUserDefaults standardUserDefaults] objectForKey:[NSString stringWithFormat:@"lastUpdateForNode%@", _destFileName]];
        if (lastUpdated && [lastUpdated length] != 0) {
            [_lastUpdateString setString:lastUpdated];
            _lastUpdateDate = [[NSDate dateWithNaturalLanguageString:lastUpdated] retain];
            [_shortLastUpdateString setString:[_lastUpdateDate descriptionWithCalendarFormat:@"%m/%d @ %I:%M%p"]];
        } else {
            [_lastUpdateString setString:@""];
            [_shortLastUpdateString setString:@""];
            _lastUpdateDate = [[NSCalendarDate calendarDate] retain];
        }
        
        _autoUpdates        = [[NSUserDefaults standardUserDefaults] integerForKey:[NSString stringWithFormat:@"autoUpdateNode%@", _destFileName]];
        _autoUpdateRate     = [[NSUserDefaults standardUserDefaults] integerForKey:[NSString stringWithFormat:@"autoUpdateRateNode%@", _destFileName]];
        _downloadsAtStartup = [[NSUserDefaults standardUserDefaults] integerForKey:[NSString stringWithFormat:@"downloadAtStartupNode%@", _destFileName]];
        
        if ([self fileDownloaded]) {
            [_statusMessage setString:@""];
            [_longStatusMessage setString:@""];
        } else {
            [_statusMessage setString:@"!"];
            [_longStatusMessage setString:@"Press \"Reload\" to begin"];
        }
        
        [self calculateNextUpdate];
        [self resetDownload];
    }
    
    return self;
}

- (id)childAtIndex:(int)i
{
    /* NODES HAVE NO CHILDREN */
    return nil;
}

- (BOOL)isNode
{
    return YES;
}

- (BOOL)expandable
{
    return NO;
}

- (void)cleanup
{
    [[NSUserDefaults standardUserDefaults] setObject:_lastUpdateString forKey:[NSString stringWithFormat:@"lastUpdateForNode%@", 
                                                                        [_dict objectForKey:@"DestFileName"]]];
}

- (void)dealloc
{
    [_dataFolderPath release];
    [_filePath release];
    [_destFileName release];
    [_sourceURL release];
    [_itemName release];
    [_itemDescription release];
    [_nodeTitle release];
    [_lastUpdateString release];
    [_lastUpdateDate release];
    [_nextUpdateString release];
    [_nextUpdateDate release];
    [_statusMessage release];
    [_longStatusMessage release];
    
    [_download release];
    _download = nil;
    
    [_receivedData release];
    
    [_dict release];
    _dict = nil;
    
    [super dealloc];
}






#pragma mark
#pragma mark DATA ACCESS METHODS
#pragma mark

- (void)setStatusMessage:(NSString *)shortMessage long:(NSString *)longMessage
{
    [_statusMessage setString:shortMessage];
    [_longStatusMessage setString:longMessage];
    if (_delegate) {
        if ([_delegate respondsToSelector:@selector( nodeHasMessage: )] ) [_delegate nodeHasMessage:self];
    }
}

- (NSString *)itemName
{
    return _itemName;
}

- (NSString *)itemStatus
{
    return _statusMessage;
}

- (NSString *)longStatus
{
    return _longStatusMessage;
}

- (NSString *)itemDescription
{
    if (_itemDescription != nil) {
        return _itemDescription;
    } else {
        return nil;
    }
}

- (NSString *)nodeTitle
{
    return _nodeTitle;
}

- (NSString *)dataSource
{
    return [_dict objectForKey:@"DataSource"];
}

- (NSString *)URL
{
    return [_dict objectForKey:@"URL"];
}

- (NSString *)lastUpdate
{
    if (_lastUpdateString && [_lastUpdateString length] != 0) {
        return _lastUpdateString;
    } else {
        return @"Unknown";
    }
}

- (NSString *)nextUpdate
{
    if (_autoUpdates && _nextUpdateString) {
        return _nextUpdateString;
    } else {
        return @"";
    }
}

- (NSString *)shortLastUpdate
{
    if (_shortLastUpdateString && [_shortLastUpdateString length] != 0) {
        return _shortLastUpdateString;
    } else {
        return @"Unknown";
    }
}

- (NSString *)shortNextUpdate
{
    if (_autoUpdates && _shortNextUpdateString) {
        return _shortNextUpdateString;
    } else {
        return @"Not Scheduled";
    }
}

- (NSString *)filePath
{
    return _filePath;
}

- (NSString *)fileName
{
    return _destFileName;
}

- (int)fileType
{
    return _fileType;
}

- (BOOL)isDownloading
{
    return (_isDownloading);
}

- (BOOL)fileDownloaded
{
    return [[NSFileManager defaultManager] fileExistsAtPath:_filePath];
}

- (BOOL)autoUpdates
{
    return _autoUpdates;
}

- (int)autoUpdateRate
{
    return _autoUpdateRate;
}

- (BOOL)downloadsAtStartup
{
    return _downloadsAtStartup;
}











#pragma mark
#pragma mark SETTINGS ACCESS METHODS
#pragma mark

- (void)setAutoUpdates:(BOOL)autoUpdates
{
    _autoUpdates = autoUpdates;
    [[NSUserDefaults standardUserDefaults] setInteger:_autoUpdates forKey:[NSString stringWithFormat:@"autoUpdateNode%@", _destFileName]];
    [self calculateNextUpdate];
    if ([_delegate respondsToSelector:@selector( nodeDidChangeUpdateSettings: )] ) [_delegate nodeDidChangeUpdateSettings:self];
}

- (void)setAutoUpdateRate:(int)rate
{
    _autoUpdateRate = rate;
    [[NSUserDefaults standardUserDefaults] setInteger:_autoUpdateRate forKey:[NSString stringWithFormat:@"autoUpdateRateNode%@", _destFileName]];
    [self calculateNextUpdate];
    if ([_delegate respondsToSelector:@selector( nodeDidChangeUpdateSettings: )] ) [_delegate nodeDidChangeUpdateSettings:self];
}

- (void)setDownloadsAtStartup:(BOOL)downloadsAtStartup
{
    _downloadsAtStartup = downloadsAtStartup;
    [[NSUserDefaults standardUserDefaults] setInteger:_downloadsAtStartup forKey:[NSString stringWithFormat:@"downloadAtStartupNode%@", _destFileName]];
}

- (int)nextUpdateInMinutes
{
    int timeForUpdate = 0;
    
    switch(_autoUpdateRate) {
        case 0: /* Five Minutes */
            timeForUpdate = 5; /* DEBUG SET BACK TO 5 */
        break;
        
        case 1: /* Five Minutes */
            timeForUpdate = 10;
        break;
        
        case 2: /* 30 Minutes */
            timeForUpdate = 30;
        break;
        
        case 3: /* 1 Hour */
            timeForUpdate = 60;
        break;
        
        case 4: /* 2 Hours */
            timeForUpdate = 120;
        break;
        
        case 5: /* 6 Hours */
            timeForUpdate = 360;
        break;
        
        case 6: /* 12 Hours */
            timeForUpdate = 720;
        break;
        
        case 7: /* 24 hours */
            timeForUpdate = 1440;
        break;
        
        default:
            timeForUpdate = 5;
        break;
    }
    return timeForUpdate;
}

- (void)calculateNextUpdate
{
    if (_nextUpdateDate) [_nextUpdateDate release];
    _nextUpdateDate = nil;
    
    NSCalendarDate *nextDate = [_lastUpdateDate dateByAddingYears:0 months:0 days:0 hours:0 minutes:[self nextUpdateInMinutes] seconds:0];
    [_nextUpdateString setString:[nextDate descriptionWithCalendarFormat:@"%m/%d at %I:%M %p"]];
    [_shortNextUpdateString setString:[nextDate descriptionWithCalendarFormat:@"%m/%d @ %I:%M%p"]];
    _nextUpdateDate = [nextDate retain];
}


- (void)setLastUpdateNow
{
    if (_lastUpdateDate) [_lastUpdateDate release];
    _lastUpdateDate = nil;
    _lastUpdateDate = [[NSCalendarDate calendarDate] retain];
    
    [_lastUpdateString setString:[_lastUpdateDate descriptionWithCalendarFormat:@"%A, %b %d %Y at %I:%M %p"]];
    [_shortLastUpdateString setString:[_lastUpdateDate descriptionWithCalendarFormat:@"%m/%d @ %I:%M%p"]];
}









#pragma mark
#pragma mark DOWNLOAD CONTROL METHODS
#pragma mark

- (void)removeDataFile
{
    if ([[NSFileManager defaultManager] fileExistsAtPath:_filePath]) {
        [[NSFileManager defaultManager] removeFileAtPath:_filePath handler:nil];
    }
}

- (void)resetDownload
{
    [_download release];
    _download = nil;
    [_receivedData release];
    _receivedData = nil;
    
    _bytesReceived = 0;
    _expectedContentLength = 0;
    _isDownloading = NO;
}

- (void)cancelDownloading
{
    if (_download) {
        [_download cancel];
    }

    [self setLastUpdateNow];
    [self calculateNextUpdate];
    [self resetDownload];
    [self setStatusMessage:@"" long:@""];
}

- (BOOL)checkAutoUpdate
{
    if (!_autoUpdates) return NO;
    if ([self isDownloading]) return NO;
    
    NSCalendarDate *now = [NSCalendarDate calendarDate];
    
    if (_nextUpdateDate) {
        if ([_nextUpdateDate isEqualToDate:now] || [now isEqualToDate:[now laterDate:_nextUpdateDate]]) {
            return YES;
        } else {
            return NO;
        }
    } else {
        return YES;
    }
}

- (void)beginOrCancelDownload
{
    // For Image Window Controller button
    if (_isDownloading) {
        [self cancelDownloading];
    } else {
        [self beginDownloadingData];
    }
}

- (void)beginDownloadingData
{    
    if (_isDownloading) return;
    
    NSURLRequest *theRequest=[NSURLRequest requestWithURL:[NSURL URLWithString:_sourceURL]
                                cachePolicy:NSURLRequestReloadIgnoringCacheData
                                timeoutInterval:[SharedController timeoutSeconds]];

    _download       = [[NSURLConnection alloc] initWithRequest:theRequest delegate:self];
    _receivedData   = [[NSMutableData data] retain];
    
    if (_download) {
        _isDownloading = YES;
        [self setStatusMessage:@"." long:@"Connecting..."];
    } else {
        if ([_delegate respondsToSelector:@selector( nodeDidFailWithError:error: )]) {
            [_delegate nodeDidFailWithError:self error:[NSString stringWithFormat:@"Unable to allocate enough memory to download data for \"%@\"", _itemName]];
        }
        [self setStatusMessage:@"Err" long:@"Error, please see Activity Window for details"];
        [self resetDownload];
    }
}



#pragma mark
#pragma mark NSURLDownload DELEGATE METHODS
#pragma mark

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
    NSString *errorMsg = [NSString stringWithFormat:@"Download of \"%@\" Failed: %@ %@",
                            _itemName, 
                            [error localizedDescription],
                            [[error userInfo] objectForKey:NSErrorFailingURLStringKey]];

    if ([_delegate respondsToSelector:@selector( nodeDidFailWithError:error: )] ) [_delegate nodeDidFailWithError:self error:errorMsg];
    [self setLastUpdateNow];
    [self calculateNextUpdate];
    [self setStatusMessage:@"Err" long:@"Error, please see Activity Window for details"];
    [self resetDownload];
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
    _expectedContentLength = [response expectedContentLength];
    [_receivedData setLength:0];
    [self setStatusMessage:@"..."  long:@"Connected..."];
}

- (NSURLRequest *)connection:(NSURLConnection *)connection 
            willSendRequest:(NSURLRequest *)request
            redirectResponse:(NSURLResponse *)redirectResponse
{
    NSLog(@"Redirect request rejected");
    return nil;
}

-(NSCachedURLResponse *)connection:(NSURLConnection *)connection 
            willCacheResponse:(NSCachedURLResponse *)cachedResponse
{
    return nil;
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{    
    if (_expectedContentLength > 0) {
        _bytesReceived += [data length];
        double percentComplete = ((double)_bytesReceived / (double)_expectedContentLength) * 100;
        [self setStatusMessage:[NSString stringWithFormat:@"%d%%",(int)percentComplete]  long:[NSString stringWithFormat:@"%d%% Complete...",(int)percentComplete]];
    }
    [_receivedData appendData:data];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
    [self setLastUpdateNow];
    [self calculateNextUpdate];
    
    if ([_receivedData length] > 0) {
        if (([_receivedData length] != _expectedContentLength) && !_overrideDataMismatch) {
            [self setStatusMessage:@"Err" long:@"Error, please see Activity Window for details"];
            if ([_delegate respondsToSelector:@selector( nodeDidFailWithError:error: )]) {
                [_delegate nodeDidFailWithError:self error:[NSString stringWithFormat:@"Download of \"%@\" Failed: data length does not match expected size (the file may not be available at this time)", _itemName]];
            }
        } else {
            
            if ([_receivedData writeToFile:_filePath atomically:YES]) {
                [self setStatusMessage:@"" long:@""];
                if ([_delegate respondsToSelector:@selector( nodeFileDidLoad: )]) [_delegate nodeFileDidLoad:self];
            } else {
                [self setStatusMessage:@"Err" long:@"Error, please see Activity Window for details"];
                if ([_delegate respondsToSelector:@selector( nodeDidFailWithError:error: )]) {
                    [_delegate nodeDidFailWithError:self error:[NSString stringWithFormat:@"Download of \"%@\" Failed: unable to write file", _itemName]];
                }
            }
            
        }
    } else {
        [self setStatusMessage:@"Err" long:@"Error, please see Activity Window for details"];
        if ([_delegate respondsToSelector:@selector( nodeDidFailWithError:error: )]) {
            [_delegate nodeDidFailWithError:self error:[NSString stringWithFormat:@"Download of \"%@\" Failed: data length is zero (the file may not be available at this time)", _itemName]];
        }
    }
    
    [self resetDownload];
}

@end



